/**
 * This file is part of the "Deniable Encryption" release in support of the
 * Blackhat Asia 2014 talk:
 *    A. Trachtenberg, "Say It Ain't So - An Implementation of Deniable Encryption".
 * 
 * The code in this file may code under patent and is released under the
 * following agreement:
 * http://nislab.bu.edu/wp-content/uploads/2014/03/Patented-Software-Release-Agreement.pdf
 * 
 * Please go to the following papers to better understand the theoretical basis of this work:
 *   A Filtser, J. Jin, A. Kontorovich, and A. Trachtenberg,
 *       "Efficient determination of the unique decodability of a string", IEEE ISIT 2013.
 *   A. Kontorovich and A. Trachtenberg,
 *       "String reconciliation with unknown edit distance", IEEE ISIT 2012.
 * 
 */

/* 
 * File:   DeBriujnGraph.cpp
 * Author: Ari Trachtenberg
 * 
 * Created on November 11, 2013, 2:05 PM
 */

// includes

#include <map>
#include <set>
#include <string>
#include <iostream>
#include <sstream>
#include "DeBruijnGraph.h"

using namespace std;

/**
 * Creates a DeBruijn multi-graph representation of a shingle set.
 * Each edge in the graph has an integer weight corresponding to its number of repetitions in the shingle set.
 * @param shingles
 */
DeBruijnGraph::DeBruijnGraph(multiset<string> shingles) {
    edgeID=0;
    totalEdges=0;
    multiset<string>::iterator it;
    string prefix, suffix;
    for (it = shingles.begin();
            it != shingles.end();
            it++) {
        prefix = it->substr(0, it->length() - 1);
        suffix = it->substr(1);
        addVert(prefix);
        addVert(suffix);
        addEdge(prefix, suffix, INIT_WEIGHT); // give each edge an initial weight
    }
}


// GRAPH QUERY

string DeBruijnGraph::print() {
    stringstream tmp;
    vtype::iterator vert_iter;
    tmp << "Vertices: ";
    for (vert_iter = Vert.begin(); vert_iter != Vert.end(); vert_iter++)
        tmp << "'" << *vert_iter << "' ";
    tmp << endl;

    etype::iterator e1;
    etype_second::iterator e2;
    tmp << "Edges:" << endl;
    for (e1 = Edge.begin(); e1 != Edge.end(); e1++)
        for (e2 = e1->second.begin(); e2 != e1->second.end(); e2++)
            tmp << "   " << e1->first << " -> " << e2->first << " (weight=" << e2->second << ")" << endl;
    tmp << endl;
    return tmp.str();

}

// GRAPH MANIPULATION

bool DeBruijnGraph::addVert(etype_v1 vertex) {
    return Vert.insert(vertex).second;
}

void DeBruijnGraph::addEdge(etype_v1 v1, etype_v2 v2, etype_weight weight) {
    totalEdges++;
    Edge[v1].insert(pair<etype_v2, etype_weight> (v2, weight));
}

string DeBruijnGraph::DFS(etype_v1 start) {
    // loosely based on CLRS
    stringstream result;
    result << start; // the first shingle    
    result << DFSvisit(start); // first visit the start node (this should traverse the whole graph if the spec requirements have been met))

    return result.str();
}

string DeBruijnGraph::DFSvisit(etype_v1 start) {
    etype_second::iterator nIter; // iterator through the neighbors of start

    stringstream result;

    // find the smallest weight edge first
    int minWt = INIT_WEIGHT;
    etype_second::iterator minVert;
    for (nIter = Edge[start].begin(); nIter != Edge[start].end(); nIter++) {
        if (nIter->second < minWt) {
            minWt = nIter->second;
            minVert = nIter;
        }
    }
    if (minWt != INIT_WEIGHT) { // there is something new to see
        minVert ->second = INIT_WEIGHT; // don't look at this edge again
        result << minVert->first.substr(minVert->first.length() - 1); // add the last character to the result
        result << DFSvisit(minVert->first);
    }
    return result.str();
}

void DeBruijnGraph::DFSupdate(etype_v1 start) {
    // loosely based on CLRS
    edgeID = -totalEdges;

    DFSupdateVisit(start); // first visit the start node

    // look for an edge that hasn't been touched, and visit it
    etype::iterator eIter;
    for (eIter = Edge.begin(); eIter != Edge.end(); eIter++) {
        etype_second::iterator nIter; // iterator through the neighbors of incident vertex
        for (nIter = eIter->second.begin(); nIter != eIter->second.end(); nIter++) {
            if (nIter->second == INIT_WEIGHT) { // I found one
                edgeID -= 2 * totalEdges; // separate different cycles by ID range
                nIter->second = edgeID++; // update its ID
                DFSupdateVisit(nIter->first); // recurse on the next vertex)
            }
        }
    }
}

void DeBruijnGraph::DFSupdateVisit(etype_v1 start) {
    etype_second::iterator nIter; // iterator through the neighbors of start
    bool found = false; // true iff we find an edge to traverse
    for (nIter = Edge[start].begin(); nIter != Edge[start].end() && !found; nIter++)
        if (nIter->second == INIT_WEIGHT) { // if this neighbor has not been visited
            found = true;
            nIter->second = edgeID++; // timestamp the visit of this edge
            DFSupdateVisit(nIter->first);
        }
}


// STATIC METHODS

multiset<string> DeBruijnGraph::shingle(string text, int shingleLen) {
    multiset<etype_v1> result;
    for (int ii = 0; ii < text.length() - shingleLen + 1; ii++) // produce the shingles
        result.insert(text.substr(ii, shingleLen));
    return result;
}

/**
 * Checks whether the length <shingleLen> shingling of the given text <text>
 * can be uniquely decodable.  Uses the Kontorovich-Trachtenberg algorithm from ITA 2013.
 * 
 * @param shingleLen The shingle length of the text
 * @return true iff the length <shingleLen> shingling of text is uniquely decodable
 */
bool DeBruijnGraph::CheckUniqDecod(const string text, int shingleLen) {
    // Create the vertex and Psi maps
    map<string, bool> Visited;
    map<string, bool> InCycle;
    map<string, string> Psi;
    map<int, string> ss;
    map<int, string> vv;
    string vi, vj;

    for (int ii = 0; ii < text.length() - shingleLen + 1; ii++) { // produce the shingles and vertices
        ss[ii] = text.substr(ii, shingleLen);
        vv[ii] = text.substr(ii, shingleLen-1);
        Visited[ss[ii]] = false;
    }
    
    for (int ii = 0; ii < text.length() - shingleLen + 1; ii++) {
        vi = vv[ii]; // v_i in the algorithm
        if (Visited[vi] == false)
            Visited[vi] = true;
        else if (InCycle[vi] == false) {
            int jj = ii;
            vj = vi;
            do {
                if (InCycle[vj] == false) {
                    InCycle[vj] = true;
                    Psi[vj] = ss[jj - 1];
                }
                jj--; vj = vv[jj];
            } while (vi != vj);
        } else { // (InCycle[vi]==true)
            //cout << ii << " - Intruding on '" << vi << "' from '" << ss[ii-1] << "'; Psi is '" << Psi[vi] << "'" << endl;
            if (Psi[vi] == ss[ii - 1])
                ;
            else
                return false;
        }
    }

    return true;
}

/**
 * Decodes a multiset of shingles into a string (i.e. finds an Eulerian cycle in the DeBruijn Graph).
 * Approach:
 * 1.  Traverse the DeBruijn graph represented by the shingles until no more new edges are available.
 *          (this should occur back at the original vertex).
 * 1b.     Timestamp each edge as it is encountered.
 * 2.  Restart at a vertex with an unvisited edge and continue with step 1.
 * 3.  When there are no unvisited edges, perform a DFS of the graph, choosing higher timestamp edges
 *         before lower timestamp edges whenever there is a choice.  Report shingles in order of DFS traversal.
 */
string DeBruijnGraph::Decode(multiset<string> shingles, string start) {
    DeBruijnGraph myGraph(shingles); // Compute DeBruijn graph
    myGraph.DFSupdate(start); // specialized DFS
    //cout << myGraph.print() << endl;
    return myGraph.DFS(start);
}