/**
 * This file is part of the "Deniable Encryption" release in support of the
 * Blackhat Asia 2014 talk:
 *    A. Trachtenberg, "Say It Ain't So - An Implementation of Deniable Encryption".
 * 
 * The code in this file may code under patent and is released under the
 * following agreement:
 * http://nislab.bu.edu/wp-content/uploads/2014/03/Patented-Software-Release-Agreement.pdf
 */

/* 
 * File:   DeniableDist.cpp
 * Author: Ari Trachtenberg
 * 
 * Created on February 2014
 */

#include <fstream>

#include "Logger.h"
#include "Auxiliary.h"
#include "Deniable.h"
#include "DataObject.h"
#include "CommString.h"
#include "SyncMethod.h"
#include "GenSync.h"
#include "CPISync_HalfRound.h"
#include "DeBruijnGraph.h"
#include "Deniable.h"

/**
 * Reads a file into a string and returns the string.
 * @param fileName The file name to read.
 * @return The contents of the file, as a string.
 */
string Deniable::readFile(string fileName) {
    string result;
    string line;
    ifstream myfile(fileName.c_str());
    if (myfile.is_open()) {
        while (getline(myfile, line))
            result += line;
        myfile.close();
    } else {
        cout << "Unable to open file " << fileName << endl;
        exit(-1);
    }
    return result;
}

/**
 * Writes <contents> to a file with name <fileName>
 * @param fileName The name of the file to which tow rite.
 * @param contents What to write to the file.
 */
void Deniable::writeFile(string fileName, string contents) {
    ofstream myfile(fileName.c_str());
    if (myfile.is_open()) {
        myfile << contents;
    } else {
        cout << "Unable to open file " << fileName << endl;
        exit(-1);
    }
    myfile.close();
}

/**
 * Helper procedure for shortestShingle()
 * @param plaintext The text to search
 * @param lower A lower bound on the shingle length needed to uniquely shingle <plaintext>
 * @param upper An upper bound on the shingle length needed to uniquely shingle <plaintext>
 * @return 
 */
int Deniable::shortestShingleHelp(const string plaintext, int lower, int upper) {
   Logger::gLog(Logger::METHOD_DETAILS, "Unique decoding shingle length search " + plaintext.substr(0, 5) + "... " + toStr(lower) + " to " + toStr(upper)); //cout << "Length " << shingleLen << ":  " << DeBruijnGraph::CheckUniqDecod(plaintext,shingleLen) << endl << "---" << endl;
   if (lower==0 || lower>=upper-1)
        return upper;
    if (DeBruijnGraph::CheckUniqDecod(plaintext,(lower+upper)/2))
        return shortestShingleHelp(plaintext,lower, (lower+upper)/2);
    else
        return shortestShingleHelp(plaintext,(lower+upper)/2, upper);
}


/**
 * Finds the shortest shingle length under which the given plaintext is uniquely decodable.
 * Uses a binary search to find the right shingle length <shingleLen>.
 * @param plaintext The plaintext to test
 * @return The shingle length.
 */
int Deniable::shortestShingle(const string plaintext) {
    int shingleLen; // first find the shortest shingling length that keeps it uniquely decodable
    for (shingleLen = 1; shingleLen < plaintext.length() && !DeBruijnGraph::CheckUniqDecod(plaintext, shingleLen); shingleLen*=2)
        Logger::gLog(Logger::METHOD, "Looking for unique decoding of " + plaintext.substr(0, 5) + "... shingle length: " + toStr(shingleLen)); //cout << "Length " << shingleLen << ":  " << DeBruijnGraph::CheckUniqDecod(plaintext,shingleLen) << endl << "---" << endl;
    return shortestShingleHelp(plaintext, shingleLen/2, shingleLen); // narrow down the search 
}

/**
 * Runs one-way CPISync to the multiset <shingles> to enable determination of up to <maxDiff>
 * differences with a probability of error <perr>.
 * @param shingles The multiset of shingles whose characteristic polynomial needs to be evaluated.
 * @param maxDiff The maximum number of differences that can be determined with the output provided.
 * @param bb The number of bits used to represent a set element
 * @param perr The maximum probability of error of reconciling up to maxDiff differences against
 *      the output provided.
 * @require <shingles> contains at least one shingle.
 * @require All shingles are of the same length, and consist of char's.
 * @return <maxDiff>\n<perr>\n<The base64 encoded output of CPISync>
 */
string Deniable::evalCPI(multiset<string> shingles, int maxDiff, int bb, double perr) {
    // initializations
    int negLogPerr = (int) -log(perr) / log(2); // the negative log of perr ... this is how the constructors expect their error value
    DataObject::RepIsInt = false; // data is in strings that should be packed

    // set up the sync elements
    // ... communicant
    list<Communicant* > clist;
    CommString* comm = new CommString("", true);
    clist.push_back(comm);

    // ... sync method
    list<SyncMethod* > mlist;
    SyncMethod *cpi_oneway = new CPISync_HalfRound(maxDiff,
            bb, // number of bits needed to represent an element
            negLogPerr);
    mlist.push_back(cpi_oneway);

    // ... data
    list<DataObject*> data;
    multiset<string>::const_iterator iter;
    for (iter = shingles.begin(); iter != shingles.end(); iter++)
        data.push_back(new DataObject(*iter));

    // ... deploy the sync
    GenSync *theSync;
    theSync = new GenSync(clist, mlist, data);
    bool syncSuccess = theSync->startSync(0);

    Logger::gLog(Logger::METHOD, "Result of encoding synchronization: " + toStr(syncSuccess));
    if (!syncSuccess)
        throw string("Encoding failed!");

    string result = comm->getString();
    string output = base64_encode(result, result.length());

    // free up memory
    delete theSync; // deletes also comm, cpi_oneway and the data elements

    // return the result, with some initial metadata
    return toStr(maxDiff) + " " + toStr(bb) + " " + toStr(perr) + " " + output;
}

/**
 * Synchronize given multiset of data A with the CPI hash of a different multiset B, returning B.
 * @param hashContents The output of a oneway CPIsync of multiset B of strings.
 *        The CPIsync must have been done with <maxDiff> differences and probability of error <perr>.
 * @param myData The contents of multiset A of strings.
 * @param maxDiff The maximum number of differences between A and B, as included in the construction of <hashContents>.
 * @param bb The number of bits per set element.
 * @param perr The maximum probability of error included in the construction of <hashContents>.
 * @require All strings in the multisets must be the same length.
 * @require There must be at least one string in myData.
 * @return The contents of multiset B.
 */
multiset<string> Deniable::syncCPI(string hashContents, const multiset<string> myData, int maxDiff, int bb, double perr) {
    // set up the sync elements
    // ... communicant
    list<Communicant* > clist;
    CommString* comm = new CommString(hashContents, true); // second param indicates that output is base64 encoded
    clist.push_back(comm);

    // ... sync method - this should really be read off "output"
    int negLogPerr = (int) -log(perr) / log(2); // the negative log of perr ... this is how the constructors expect their error value
    list<SyncMethod* > mlist2;
    mlist2.push_back(new CPISync_HalfRound(maxDiff,
            bb, // number of bits needed to represent an element
            negLogPerr));

    // ... data
    list<DataObject*> data2;
    multiset<string>::iterator iter;
    for (iter = myData.begin(); iter != myData.end(); iter++)
        data2.push_back(new DataObject(*iter));

    // ... deploy the sync
    GenSync *theSync;
    theSync = new GenSync(clist, mlist2, data2);
    bool syncSuccess = theSync->listenSync(0);

    Logger::gLog(Logger::METHOD, "Result of decoding synchronization: " + toStr(syncSuccess));
    if (!syncSuccess)
        throw string("Decoding failed!");

    // ... stitch together the results
    list<DataObject *> elems = theSync->dumpElements();
    list<DataObject *>::const_iterator iter2;
    multiset<string> shingles2;

    for (iter2 = elems.begin(); iter2 != elems.end(); iter2++) {
        //cout << (**iter2).to_string() << endl;
        shingles2.insert((**iter2).to_string());
    }

    return shingles2;
}

/**
 * Attempts to decode string <encoded> with key <key>.
 * @param ciphertext The encrypted text.
 * @param key The key being used for decoding.
 * @require key length must be the same as the shingle length used to produce <encoded>
 * @return The decoded output.
 */
string Deniable::decodeWithKey(string key, string ciphertext) {
    // ... parse initial info
    istringstream iss(ciphertext);
    int maxDiff;
    int bb;
    double perr;
    string hashContents;
    
    iss >> maxDiff >> bb >> perr >> hashContents;

    // parse and enter the key into the multiset
    Logger::gLog(Logger::METHOD, "Got key: " + key + "\n" + writeInts(key.data(), key.length()) + "\n");

    stringstream keystr(key);
    int shingleLen;
    keystr >> shingleLen; // get the shingle length
    char tmp[1];
    keystr.read(tmp, 1); // move past the end of string char
    multiset<string> myData;
    // read key shingles one by one
    string keyShingle;
    int offset = key.find(Deniable::DELIMITER_NULL); // the start of a string
    int offset2;                           // the end of a string

    do {
        offset2 = key.find(Deniable::DELIMITER_NULL, offset + 1);
        keyShingle = key.substr(offset+1, offset2 - offset-1);
        multiset<string> shingles = DeBruijnGraph::shingle(keyShingle, shingleLen);
        
        myData.insert(shingles.begin(), shingles.end()); // all the shingles produced from keyShingle
        offset = offset2;
    } while (offset2 != key.length()-1);
    multiset<string> shingles = syncCPI(hashContents, myData, maxDiff, bb, perr);

    // find the shingle that contains the null character - it's the first shingle, by definition
    //cout << "Shingles are " << multisetPrint(shingles) << endl;
    multiset<string>::iterator iter;
    for (iter = shingles.begin();
            (iter->find(Deniable::DELIMITER) == string::npos) && (iter != shingles.end());
            iter++)
        ;

    string decoded = DeBruijnGraph::Decode(shingles, iter->substr(0, iter->length() - 1)); // decode starting from the first shingle
    return decoded;
}

/**
 * (Greedily) combine strings in a given multiset of shingles where the end of the first
 * string matches the beginning of another string.  The current implementation is worst-case
 * quadratic time in the number of shingles.
 * @param shingles The shingles to attempt to combine
 * @return A collection of shingles, some of which may have been combined
 */
multiset<string> Deniable::compress(const multiset<string> shingles) {
    multiset<string> compressed = shingles;
    bool matched = true; // true iff two shingles matched in the previous iteration
    while (matched) {
        matched = false;
        multiset<string>::iterator it, it2;
        for (it = compressed.begin(); !matched && it != compressed.end(); it++) {
            for (it2 = compressed.begin(); !matched && it2 != compressed.end(); it2++) {
                if (*it == *it2) // shingles are the same
                    continue;
                int common = min(it->length(), it2->length());
                string suffix = it->substr(it->length() - common + 1); // the suffix of one string 
                string prefix = it2->substr(0, suffix.length()); // the prefix of the other
                if (suffix == prefix) { // the two substrings may be combined
                    string newShingle = (*it) + it2->substr(prefix.length()); // concatenate it and it2 (without doubling overlap region)
                    compressed.erase(it);
                    compressed.erase(it2);
                    compressed.insert(newShingle);
                    matched = true;
                    break;
                }
            }
            if (matched) break; // i.e. don't iterate a possible unusable iterator
        }
    }
    return compressed;
}

/**
 * Computes a key for a given collection of shingles (the ciphertext is computed elsewhere).
 * @param shingleLen The common length of each shingle.
 * @param shingles The shingles to include in the key.
 * @requires Maintains the assumption that \0 does not appear within any shingle
 * @return A string representing a valid key for the given parameters.
 */
string Deniable::computeKey(int shingleLen, const multiset<string> &shingles) {
    string key = "";

    // ... write shingleLen to the key
    key += toStr(shingleLen) + toStr(Deniable::DELIMITER_NULL);

    // compress shingles greedily
    multiset<string> compressed = compress(shingles);

    // ... add substrings from <compressed> to the key
    multiset<string>::iterator iter = compressed.begin();
    for (iter = compressed.begin(); iter != compressed.end(); iter++)
        key += *iter + toStr(Deniable::DELIMITER_NULL);

    return base64_encode(key.data(), key.length());
}

  pair< string, vector<string> > Deniable::encrypt(const vector<string> &plaintexts, double perr, int sec)  {
      string ciphertext; /** Maintains the encrypted ciphertext. */
      
      // 1.  Find the maximum shingle size needed to guarantee unique decoding for all input files
        int shingleLen = -1;
        // ... take max with all plaintexts
        for (int ii = 0; ii < plaintexts.size(); ii++) {
            int plainLen = shortestShingle(Deniable::DELIMITER + readFile(plaintexts[ii])); // add input delimiter for each plaintext
            if (plainLen > shingleLen)
                shingleLen = plainLen;
        }

        // ... compute intersection of all shingles
        vector< multiset<string> > shingles; // will store shingles for the plaintext files - this can be computed serially to avoid large memory usage
        shingles.push_back(DeBruijnGraph::shingle(DELIMITER + readFile(plaintexts[0]), shingleLen));
        multiset<string> intersectShingles = shingles[0]; /** Intersection of shingle sets for plaintexts. */

        for (int ii = 1; ii < plaintexts.size(); ii++) {
            shingles.push_back(DeBruijnGraph::shingle(DELIMITER + readFile(plaintexts[ii]), shingleLen)); // add the identifier fore the start of the string
            multiset<string> tmp = multisetIntersect(intersectShingles, shingles[ii]);
            intersectShingles = tmp;
        }
        // ... compute the maximum difference from the multisets to the intersection
        int maxDiff = -1;
        for (int ii = 0; ii < plaintexts.size(); ii++) {
            int tmpDiff = multisetDiff(shingles[ii], intersectShingles).size();
            if (tmpDiff > maxDiff)
                maxDiff = tmpDiff;
        }

        // 2.  Compute the encryption
        int bb = 8 * sizeof (char)*(*intersectShingles.begin()).length(); // number of bits per shingle
        // ... display parameters
        Logger::gLog(Logger::METHOD_DETAILS,"error probability: " + toStr(perr) + ")\n");
        Logger::gLog(Logger::METHOD_DETAILS,"(encoding bits per shingle: " + toStr(bb) + ")\n");

        if (sec < 0) // i.e. unset
            sec = maxDiff + MIN_SEC / bb + 1; // minimum security of an encryption
        ciphertext = evalCPI(intersectShingles, maxDiff + intersectShingles.size() - sec, bb, perr);

        // 4. Compute the key
        vector<string> keys; /** The i-th key with the ciphertext produces the i-th plaintext. */

        if (intersectShingles.size() < sec)
            throw Deniable::EncryptionFailure("Cannot meet security parameter "+toStr(sec)+".  "+
                    "Please lower security requirements or increase plaintext length.");
        for (int ii = 0; ii < plaintexts.size(); ii++) {
            keys.push_back( computeKey(shingleLen,
                    multisetUnion(
                    multisetDiff(shingles[ii], intersectShingles),
                    multisetSubset(intersectShingles, sec)))
            );
            Logger::gLog(Logger::METHOD_DETAILS, + "key " + toStr(ii) + " is:    \"" + keys[ii] + "\"\n");
       }
        
            // return the result
        return pair<string, vector<string> >(ciphertext,keys);
}

  string Deniable::decrypt(const string key, const string ciphertext) {
      try {
            return decodeWithKey(base64_decode(key), readFile(ciphertext));
        } catch (string str) {
            throw Deniable::DecryptionFailure(key,ciphertext);
        }
  }