/*
 * This code is part of the CPISync project developed at Boston University:
 *      {@link http://nislabcpi.bu.edu/cpisync}
 * 
 * The code, explanation, references, API, and a demo can be found on this
 * web page.  The main theoretical basis for the approach can be found at
 * {@link http://ipsit.bu.edu/documents/ieee-it3-web.pdf}
 *   Y. Minsky, A. Trachtenberg, and R. Zippel,
 *   "Set Reconciliation with Nearly Optimal Communication Complexity",
 *   IEEE Transactions on Information Theory, 49:9.
 * 
 * Elements of the CPISync project code have been worked on, at various points, by:
 * @author Ari Trachtenberg
 * @author Sachin Agarwal 
 * @author Paul Varghese
 * @author Jiaxi Jin
 * @author Jie Meng
 * @author Alexander Smirnov
 * @version 2.01, 11/20/2013
 */
#include <iostream>
#include <fstream>

#include "GenSync.h"
#include "Exceptions.h"
#include "Communicant.h"
#include "DataObject.h"
#include "SyncMethod.h"
#include "Logger.h"

using namespace std;

/**
 * Construct a default GenSync object - communicants and objects will have to be added later
 */
GenSync::GenSync() {
}

/**
 * Construct a specific GenSync object
 */
GenSync::GenSync(const list<Communicant*> clist, const list<SyncMethod*> mlist, const list<DataObject*> data) {
    myCommList = clist;
    mySyncList = mlist;
    outFile = NULL; // no output file is being used

    // add each datum one by one
    list<DataObject*>::const_iterator itData = data.begin();
    for (; itData != data.end(); itData++)
        addElem(*itData);
}

GenSync::GenSync(const list<Communicant*> clist, const list<SyncMethod*> mlist, string fileName) {
    myCommList = clist;
    mySyncList = mlist;
    outFile = NULL; // add elements without writing to the file at first
    Logger::gLog(Logger::METHOD,"Entering GenSync::GenSync");
    // read data from a file
    Logger::gLog(Logger::METHOD, "Utilizing file: " + fileName);
    ifstream inFile(fileName.c_str());
    string str;
    for (getline(inFile, str); inFile.good(); getline(inFile, str)) {
        addElem(new DataObject(str)); // add this datum to our list
        Logger::gLog(Logger::METHOD_DETAILS, "... added set element " + str);
    }
    inFile.close();

    // register the file to which new data should be appended
    outFile = new ofstream(fileName.c_str(), ios::app);

}

// destruct a gensync object

GenSync::~GenSync() {
    // clear out memory
    myData.clear();

    list<SyncMethod*>::iterator itAgt = mySyncList.begin();
    for (; itAgt != mySyncList.end(); itAgt++) {
        delete *itAgt;
    }
    mySyncList.clear();

    list<Communicant*>::iterator itComm = myCommList.begin();
    for (; itComm != myCommList.end(); itComm++) {
        delete *itComm;
    }
    myCommList.clear();

    // close and free the output file
    if (outFile != NULL) {
        outFile->close();
        delete outFile;
    }

}


// listen, receive data and conduct synchronization

bool GenSync::listenSync(int method_num) {
    Logger::gLog(Logger::METHOD,"Entering GenSync::listenSync");
    // find the right syncAgent	
    list<SyncMethod*>::iterator syncAgent = mySyncList.begin();
    advance(syncAgent, method_num);

    bool syncSuccess = true; // true if all syncs so far were successful

    // ask each communicant to listen, one by one
    list<Communicant*>::iterator itComm;
    list<DataObject*> selfMinusOther, otherMinusSelf;
    for (itComm = myCommList.begin(); itComm != myCommList.end(); ++itComm) {
        // initialize variables
        selfMinusOther.clear();
        otherMinusSelf.clear();

        try {
            syncSuccess &= (*syncAgent)->SyncServer(*itComm, selfMinusOther, otherMinusSelf);
        } catch (SyncFailureException s) {
            Logger::error_and_quit(s.what());
            return false;
        }

        // add any items that were found in the reconciliation
        list<DataObject*>::iterator itDO;
        for (itDO = otherMinusSelf.begin(); itDO != otherMinusSelf.end(); itDO++)
            addElem(*itDO);
    }

    return syncSuccess;
}

// request connection, send data and get the result

bool GenSync::startSync(int method_num) {
    Logger::gLog(Logger::METHOD,"Entering GenSync::startSync");
    SyncMethod* tmpSync;
    // find the right syncAgent	
    list<SyncMethod*>::iterator syncAgent = mySyncList.begin();
    advance(syncAgent, method_num);

    bool syncSuccess = true; // true if all syncs so far were successful
    list<Communicant*>::iterator itComm;
    list<DataObject*> selfMinusOther, otherMinusSelf;
    for (itComm = myCommList.begin(); itComm != myCommList.end(); ++itComm) {
        // initialize variables
        selfMinusOther.clear();
        otherMinusSelf.clear();

        // do the sync
        try {
            if (!(*syncAgent)->SyncClient(*itComm, selfMinusOther, otherMinusSelf)) {
                Logger::gLog(Logger::METHOD,"Sync to " + (*itComm)->getName() + " failed!");
                syncSuccess = false;
            }
        } catch (SyncFailureException s) {
            Logger::error_and_quit(s.what());
            return false;
        }

        // add any items that were found in the reconciliation
        list<DataObject*>::iterator itDO;
        for (itDO = otherMinusSelf.begin(); itDO != otherMinusSelf.end(); itDO++)
            addElem(*itDO);
    }

    Logger::gLog(Logger::METHOD, "Sync succeeded:  " + toStr(syncSuccess));
    return syncSuccess;

}

// add element

void GenSync::addElem(DataObject* newDatum) {
    Logger::gLog(Logger::METHOD,"Entering GenSync::addElem");
    // store locally
    myData.push_back(newDatum);

    // update synch methods' metadata
    list<SyncMethod*>::iterator itAgt;
    for (itAgt = mySyncList.begin(); itAgt != mySyncList.end(); ++itAgt) {
        if (!(*itAgt)->addElem(newDatum))
            Logger::error_and_quit("Could not add item " + newDatum->to_string() + ".  Please considering increasing the number of bits per set element.");
    }

    // update file
    if (outFile != NULL)
        (*outFile) << newDatum->to_string() << endl;
}

template <typename T>
void GenSync::addElem(T* newDatum) {
    Logger::gLog(Logger::METHOD,"Entering GenSync::addElem");
    DataObject *newDO = new DataObject(*newDatum);
    addElem(newDO);
    delete newDO;
}

// delete element

void GenSync::delElem(DataObject* newDatum) {
    throw new UnimplementedMethodException("GenSync::delElem");
}



// insert a communicant in the list at the index position

void GenSync::addComm(Communicant* newComm, int index) {
    Logger::gLog(Logger::METHOD,"Entering GenSync::addComm");
    list<Communicant*>::iterator itComm;

    itComm = myCommList.begin();
    if (index == 0) {
        myCommList.push_back(newComm);
    } else {
        advance(itComm, index - 1);
        myCommList.insert(itComm, newComm);
    }

}

// delete a communicant from the list with name oldComm 

void GenSync::delComm(Communicant* oldComm) {
    myCommList.remove(oldComm);

}

// delete a communicant from the list at the index position

void GenSync::delComm(int index) {
    list<Communicant*>::iterator itComm;

    itComm = myCommList.begin();
    if (index == 1) {
        myCommList.erase(itComm);
    } else {
        advance(itComm, index - 1);
        myCommList.erase(itComm);
    }

}

int GenSync::numComm() {
    return myCommList.size();
}

// insert a syncmethod in the list at the index position

void GenSync::addSyncAgt(SyncMethod* newAgt, int index) {
    Logger::gLog(Logger::METHOD,"Entering GenSync::addSyncAgt");
    // create and populate the new agent
    list<DataObject*>::iterator itData;
    for (itData = myData.begin(); itData != myData.end(); itData++)
        if (!newAgt->addElem(*itData))
            Logger::error_and_quit("Was not able to add an item to the next syncagent.");

    // add the agent to the sync agents list
    mySyncList.push_back(newAgt);
}

// delete a syncmethod from the list with name oldAgt 

void GenSync::delSyncAgt(SyncMethod* oldAgt) {
    mySyncList.remove(oldAgt);
}

// delete a syncmethod from the list at the index position

void GenSync::delSyncAgt(int index) {
    list<SyncMethod*>::iterator itAgt;

    itAgt = mySyncList.begin();
    advance(itAgt, index - 1);

    mySyncList.erase(itAgt);
}

const list<DataObject *> GenSync::dumpElements() {
    return myData;
}